------------------------------------------------------------------------------------------------------------
--
-- Copyright (C) 2010 Nikolaos Ch. Alachiotis
--
-- Engineer: 			Nikolaos Alachiotis 
-- Contact:				n.alachiotis@gmail.com
-- 						http://wwwkramer.in.tum.de/exelixis/nikos
--
-- Create Date:    	05/20/2010 			 
-- Design Name: 	 	DP_LAU_DUAL
-- Module Name:    	DP_LAU_DUAL - Behavioral 
-- Project Name:   	Double-Precision Pipelined Logarithm Approximation Unit for FPGAs (Dual)
-- Target Devices: 	Virtex 5 family
-- Tool versions: 	Xilinx ISE Design Suite 10.1
--
-- Description: 		Implementation of the ICSILog algorithm.
--							A detailed description and evaluation of the LAU architecture can be found at
--							http://wwwkramer.in.tum.de/exelixis/nikos/publications.html
--
-- Evaluation:			Maximum Clock Frequency: 330 MHz (Power Optimization Strategy)
--															 329 MHz (Balanced Strategy)
--							Latency in clock cycles: 22
--							Throughput:					 2 DP Results/Cycle
--							Accuracy: 					 0.57 * 2^(-17) (Avg Error)
--
-- Additional Info:	The unit has been verified on Virtex5 SX95T-1 FPGA.
--							Use the "Power Optimization Strategy" for slightly better performance.  	
--
------------------------------------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity DP_LAU_DUAL is
    Port ( rst : in  STD_LOGIC;
           clk : in  STD_LOGIC;
           vld_in_1 : in  STD_LOGIC;
           val_in_1 : in  STD_LOGIC_VECTOR (63 downto 0);
           vld_out_1 : out  STD_LOGIC;
           val_out_1 : out  STD_LOGIC_VECTOR (63 downto 0);
			  vld_in_2 : in  STD_LOGIC;
           val_in_2 : in  STD_LOGIC_VECTOR (63 downto 0);
           vld_out_2 : out  STD_LOGIC;
           val_out_2 : out  STD_LOGIC_VECTOR (63 downto 0));
end DP_LAU_DUAL;

architecture Behavioral of DP_LAU_DUAL is

component dp_fp_log_v2 is
    Port ( rst : in  STD_LOGIC;
           clk : in  STD_LOGIC;
           valid_in : in  STD_LOGIC;
           input_val : in  STD_LOGIC_VECTOR (63 downto 0);
           valid_out : out  STD_LOGIC;
           output_val : out  STD_LOGIC_VECTOR (63 downto 0);
			  mant_lut_addr: out STD_LOGIC_VECTOR(11 downto 0);
			  mant_lut_data: in STD_LOGIC_VECTOR(26 downto 0);
			  exp_addr: out std_logic_vector(9 downto 0);
			  exp_data: in std_logic_Vector(12 downto 0)
		);
end component;

component mant_lut_dual is
  port (
    clka : in STD_LOGIC := 'X'; 
    clkb : in STD_LOGIC := 'X'; 
    addra : in STD_LOGIC_VECTOR ( 11 downto 0 ); 
    addrb : in STD_LOGIC_VECTOR ( 11 downto 0 ); 
    douta : out STD_LOGIC_VECTOR ( 26 downto 0 ); 
    doutb : out STD_LOGIC_VECTOR ( 26 downto 0 ) 
  );
end component;

component exp_lut_MEM_DUAL is
  port (
    clka : in STD_LOGIC := 'X'; 
    clkb : in STD_LOGIC := 'X'; 
    addra : in STD_LOGIC_VECTOR ( 9 downto 0 ); 
    addrb : in STD_LOGIC_VECTOR ( 9 downto 0 ); 
    douta : out STD_LOGIC_VECTOR ( 12 downto 0 ); 
    doutb : out STD_LOGIC_VECTOR ( 12 downto 0 ) 
  );
end component;

signal lau1_addr, lau2_addr: std_logic_Vector(11 downto 0);
signal lau1_data, lau2_data: std_logic_Vector(26 downto 0);
signal lau1_exp_addr, lau2_exp_addr: std_logic_Vector(9 downto 0);
signal lau1_exp_data, lau2_exp_data: std_logic_Vector(12 downto 0);

signal rst_r, vld_in_1_r, vld_in_2_r, vld_out_1_r, vld_out_2_r: std_logic;
signal val_in_1_r,val_in_2_r, val_out_1_r, val_out_2_r: std_logic_Vector(63 downto 0);

component shift_reg_32_1c is
  port (
    clk : in STD_LOGIC := 'X'; 
    d : in STD_LOGIC_VECTOR ( 31 downto 0 ); 
    q : out STD_LOGIC_VECTOR ( 31 downto 0 ) 
  );
end component;

component shift_reg_65_1c is
  port (
    clk : in STD_LOGIC := 'X'; 
    d : in STD_LOGIC_VECTOR ( 64 downto 0 ); 
    q : out STD_LOGIC_VECTOR ( 64 downto 0 ) 
  );
end component;

signal lau1_inv,lau1_inv_r,lau1_outv,lau1_outv_r: std_logic_Vector(64 downto 0);
signal lau2_inv,lau2_inv_r,lau2_outv,lau2_outv_r: std_logic_Vector(64 downto 0);

begin

lau1_inv(64)<=vld_in_1;
lau1_inv(63 downto 0)<=val_in_1;
LAU1_SFHT_IN1: shift_reg_65_1c port map(clk,lau1_inv,lau1_inv_r);
vld_out_1<=lau1_outv_r(64);
val_out_1<=lau1_outv_r(63 downto 0);
LAU1_SFHT_IN2: shift_reg_65_1c port map(clk,lau1_outv,lau1_outv_r);


lau2_inv(64)<=vld_in_2;
lau2_inv(63 downto 0)<=val_in_2;
LAU2_SFHT_IN1: shift_reg_65_1c port map(clk,lau2_inv,lau2_inv_r);
vld_out_2<=lau2_outv_r(64);
val_out_2<=lau2_outv_r(63 downto 0);
LAU2_SFHT_IN2: shift_reg_65_1c port map(clk,lau2_outv,lau2_outv_r);

LAU_1: dp_fp_log_v2 Port Map
( rst => rst_r,
  clk => clk,
  valid_in => lau1_inv_r(64),
  input_val => lau1_inv_r(63 downto 0),
  valid_out => lau1_outv(64),
  output_val => lau1_outv(63 downto 0),
  mant_lut_addr => lau1_addr,
  mant_lut_data => lau1_data,
  exp_addr => lau1_exp_addr,
  exp_data => lau1_exp_data
);

MANT_LUT: mant_lut_dual Port Map
(
    clka => clk, 
	 clkb => clk,
    addra => lau1_addr, 
	 addrb => lau2_addr,
    douta => lau1_data,
	 doutb => lau2_data	 
);
  
EXP_LUT: exp_lut_MEM_DUAL Port Map
(
    clka => clk, 
	 clkb => clk, 
    addra => lau1_exp_addr, 
    addrb => lau2_exp_addr,
    douta => lau1_exp_data, 
    doutb => lau2_exp_data
);

LAU_2: dp_fp_log_v2 Port Map
( rst => rst_r,
  clk => clk,
  valid_in => lau2_inv_r(64),
  input_val => lau2_inv_r(63 downto 0),
  valid_out => lau2_outv(64),
  output_val => lau2_outv(63 downto 0),
  mant_lut_addr => lau2_addr,
  mant_lut_data => lau2_data,
  exp_addr => lau2_exp_addr,
  exp_data => lau2_exp_data
);

end Behavioral;

