------------------------------------------------------------------------------------------------------------
--
-- Copyright (C) 2010 Nikolaos Ch. Alachiotis
--
-- Engineer: 			Nikolaos Alachiotis 
-- Contact:				n.alachiotis@gmail.com
-- 						http://wwwkramer.in.tum.de/exelixis/nikos
--
-- Create Date:    	05/20/2010 			 
-- Design Name: 	 	SP_LAU_DUAL
-- Module Name:    	SP_LAU_DUAL - Behavioral 
-- Project Name:   	Single-Precision Pipelined Logarithm Approximation Unit for FPGAs (Dual)
-- Target Devices: 	Virtex 5 family
-- Tool versions: 	Xilinx ISE Design Suite 10.1
--
-- Description: 		Implementation of the ICSILog algorithm.
--							A detailed description and evaluation of the LAU architecture can be found at
--							http://wwwkramer.in.tum.de/exelixis/nikos/publications.html
--
-- Evaluation:			Maximum Clock Frequency: 351 MHz (Power Optimization Strategy)
--															 332 MHz (Balanced Strategy)
--							Latency in clock cycles: 22
--							Throughput:					 2 SP Results/Cycle
--							Accuracy: 					 0.57 * 2^(-17) (Avg Error)
--
-- Additional Info:	The unit has been verified on Virtex5 SX95T-1 FPGA.
--							Use the "Power Optimization" optimization strategy for slightly better performance.  	
--
------------------------------------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity SP_LAU_DUAL is
    Port ( rst : in  STD_LOGIC;
           clk : in  STD_LOGIC;
           vld_in_1 : in  STD_LOGIC;
           val_in_1 : in  STD_LOGIC_VECTOR (31 downto 0);
           vld_out_1 : out  STD_LOGIC;
           val_out_1 : out  STD_LOGIC_VECTOR (31 downto 0);
			  vld_in_2 : in  STD_LOGIC;
           val_in_2 : in  STD_LOGIC_VECTOR (31 downto 0);
           vld_out_2 : out  STD_LOGIC;
           val_out_2 : out  STD_LOGIC_VECTOR (31 downto 0));
end SP_LAU_DUAL;

architecture Behavioral of SP_LAU_DUAL is

component sp_fp_log_v2 is
    Port ( rst : in  STD_LOGIC;
           clk : in  STD_LOGIC;
           valid_in : in  STD_LOGIC;
           input_val : in  STD_LOGIC_VECTOR (31 downto 0);
           valid_out : out  STD_LOGIC;
           output_val : out  STD_LOGIC_VECTOR (31 downto 0);
			  exp_addr: out  STD_LOGIC_VECTOR(6 downto 0);
			  exp_data: in  STD_LOGIC_VECTOR(8 downto 0);
			  mant_lut_addr: out STD_LOGIC_VECTOR(11 downto 0);
			  mant_lut_data: in  STD_LOGIC_VECTOR(26 downto 0));
end component;

component mant_lut_MEM is
  port (
    clka : in STD_LOGIC := 'X'; 
    addra : in STD_LOGIC_VECTOR ( 11 downto 0 ); 
    douta : out STD_LOGIC_VECTOR ( 26 downto 0 ) 
  );
end component;

component mant_lut_dual is
  port (
    clka : in STD_LOGIC := 'X'; 
    clkb : in STD_LOGIC := 'X'; 
    addra : in STD_LOGIC_VECTOR ( 11 downto 0 ); 
    addrb : in STD_LOGIC_VECTOR ( 11 downto 0 ); 
    douta : out STD_LOGIC_VECTOR ( 26 downto 0 ); 
    doutb : out STD_LOGIC_VECTOR ( 26 downto 0 ) 
  );
end component;

component exp_lut_MEM_DUAL is
  port (
    clka : in STD_LOGIC := 'X'; 
    clkb : in STD_LOGIC := 'X'; 
    addra : in STD_LOGIC_VECTOR ( 6 downto 0 ); 
    addrb : in STD_LOGIC_VECTOR ( 6 downto 0 ); 
    douta : out STD_LOGIC_VECTOR ( 8 downto 0 ); 
    doutb : out STD_LOGIC_VECTOR ( 8 downto 0 ) 
  );
end component;

signal lau1_addr, lau2_addr: std_logic_Vector(11 downto 0);
signal lau1_data, lau2_data: std_logic_Vector(26 downto 0);
signal lau1_exp_addr, lau2_exp_addr: std_logic_Vector(6 downto 0);
signal lau1_exp_data, lau2_exp_data: std_logic_Vector(8 downto 0);

begin

LAU_1: sp_fp_log_v2 Port Map
( rst => rst,
  clk => clk,
  valid_in => vld_in_1,
  input_val => val_in_1,
  valid_out => vld_out_1,
  output_val => val_out_1,
  exp_addr => lau1_exp_addr,
  exp_data => lau1_exp_data,
  mant_lut_addr => lau1_addr,
  mant_lut_data => lau1_data
);

MANT_LUT: mant_lut_dual Port Map
(
    clka => clk, 
	 clkb => clk,
    addra => lau1_addr, 
	 addrb => lau2_addr,
    douta => lau1_data,
	 doutb => lau2_data	 
);
  
EXP_LUT: exp_lut_MEM_DUAL Port Map
(
    clka => clk, 
	 clkb => clk, 
    addra => lau1_exp_addr, 
    addrb => lau2_exp_addr,
    douta => lau1_exp_data, 
    doutb => lau2_exp_data
);
  
LAU_2: sp_fp_log_v2 Port Map
( rst => rst,
  clk => clk,
  valid_in => vld_in_2,
  input_val => val_in_2,
  valid_out => vld_out_2,
  output_val => val_out_2,
  exp_addr => lau2_exp_addr,
  exp_data => lau2_exp_data,
  mant_lut_addr => lau2_addr,
  mant_lut_data => lau2_data
);


end Behavioral;

